// ******************************************************************************************************
// BSP : STM32 MCU based boards by STMicroelectronics 2.7.1
// Borad select : STM32 based boards / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number')
// ******************************************************************************************************
// 1. wait for transmission of Slave
// 2. receives Slave data 11 bytes(SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
// 3. transtits response data 11 bytes(MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)

// 2024/07/07 @msfujino

#include <Adafruit_GFX.h>       // 1.11.9
#include <Adafruit_SSD1306.h>   // 2.5.10
#include "myFonts/tahoma7pt7b.h"  // any font
#include <RadioLib.h>           // 6.6.0  RadioLib by Jan Gromes (Library manager)
#include <SPI.h>
#include <SD.h>

#define CHECK_INTERVAL 20000    // communication check interval [mS]

// BAT+_pin() --> 1000k --> A4_pin --> 510k --> D0_pin
#define PIN_VBATT      A4       // battery voltage measurement pin
#define READ_ENABLE    D0       // read PIN_VBATT enable LOW:enable, HIGH:disable

#define SD_CS         PB9       // SD chip select pin

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
// Serial.println(*((uint32_t*)0x1FFF7580));
#define MASTER_UID  *((uint32_t*)0x1FFF7580)    // use uid64 for Master device UID
#define SLAVE_UID   0x05370378  // example Slave device UID

Adafruit_SSD1306 display(128, 64, &Wire, -1);   // OLED display
File logFile;                                   // SD log file

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] = {PA4, PA5, RADIOLIB_NC, RADIOLIB_NC, RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
//  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini
  END_OF_MODE_TABLE,
};

#define dataNum   11                // receive data 11 bytes  (SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
                                    // transmit data 11 bytes (MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
union unionBuff {                   // buffer for data type conversion
  uint32_t  Buff_u32[dataNum/4];    // SlaveID, MasterID, randomNumber, verifyNumber
  uint16_t  Buff_u16[dataNum/2];    // Vbatt
  int8_t    Buff_i8[dataNum];       // rssi, snr
  uint8_t   Buff_u8[dataNum];       // verifyResult, interval
};
union unionBuff ub;

int8_t rssi;                        // signal RSSI [dBm]
int8_t snr;                         // signal SN ratio [dB]
uint32_t masteruid = MASTER_UID;    // Master device UID
uint32_t slaveuid;                  // Slave device UID
uint16_t Vbatt;                     // battery voltage data [V]
uint32_t randomNumber;              // random number to verify
uint32_t verifyNumber;              // verify number
uint8_t verifyResult;               // received verify result
char printBuff[4];                  // for sprintf()
String  txdata = "";                // transmission data packet string
String  rxdata = "";                // received data packet string
bool rxstatus = false;              // receive status flag
bool receivedFlag = false;          // flag that a packet was received
bool transmittedFlag = false;       // flag that a packet was transmitted
String logFileName = "LoRaLog.txt"; // Log file name
uint32_t timestamp = millis();      // previouse received time [mS]
uint32_t receivedTime = millis();   // packet received time [mS]
uint8_t receivedInterval;           // packet received interval [sec]

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
//  while(!Serial);
  delay(2000);

  // Master UID
  Serial.print("Master UID "); Serial.println(*((uint32_t*)0x1FFF7580), HEX);

  // Pin initialization
  pinMode(LED_RED, OUTPUT);         // built-in LED
  pinMode(READ_ENABLE, OUTPUT);     // read PIN_VBATT enable

  // Display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Master");  
  display.display();
  delay(2000);
  display.clearDisplay();

  // SD initialization and file open
  if(!SD.begin(SD_CS)) {                     // SCK 4MHz   
    Serial.println("SD begin failure!");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("SD begin failure");
    display.display();
    errorBlink(2);                           // error [[ 2 ]]
  } 

  logFile = SD.open(logFileName, FILE_WRITE);
  if(!logFile) { 
    Serial.println(logFileName + " open failure");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("file open failure");
    display.display();
    errorBlink(3);                           // error [[ 3 ]]   
  }   
  Serial.println("SD initialization finished");

  // STM32WL initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);


  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink(4);                           // error [[ 4 ]]
  }

  // callback function when received or transmitted
  radio.setDio1Action(setFlag);

  // setup end
  display.setCursor(0, 15);
  display.print("Data Verify"); 
  display.setCursor(0, 47);
  display.print("LoRa-E5 Master");
  display.setCursor(0, 63);
  display.print("Setup END");
  display.display();
  delay(2000);
  display.clearDisplay(); 
}

// ***********************************************************************************************************
void loop() 
{
  // Master receives packet from Slave ***********************************************************************
  Serial.println("Waiting for incoming Slave transmitting packet ... ");

  // start listening for LoRa packets
  receivedFlag = false;
  int state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink_1(2);                           // error [[ 2 ]]
  }

  // check if the flag is set
  // If not received, proceed to next communication cycle to attempt synchronization recovery
  uint32_t checkstart = millis();
  digitalWrite(LED_RED, LOW);
  while(!receivedFlag & (millis() - checkstart < CHECK_INTERVAL)) {
    delay(1);   // wait for SPI ready
  }
  digitalWrite(LED_RED, HIGH);
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ub.Buff_u8, dataNum);

  slaveuid = ub.Buff_u32[0];    // Slave device UID
  if (state == RADIOLIB_ERR_NONE & slaveuid == SLAVE_UID) {   // Check Slave ID   

    // packet form Slave was successfully received
    receivedTime = millis();
    receivedInterval = (receivedTime - timestamp) / 1000;       // [sec]
    timestamp = receivedTime;

    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ub.Buff_u8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    randomNumber = ub.Buff_u32[1];    // random number to verify
    Vbatt = ub.Buff_u16[4];           // battery voltage [mV]
    verifyResult = ub.Buff_u8[10];    // verify result 

    Serial.printf("Slave UID:\t%08x\n", slaveuid);    
    Serial.print("Interval:\t"); Serial.print(receivedInterval); Serial.println(" sec");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt); Serial.println(" mV");
    Serial.print("randomNumber:\t"); Serial.println(randomNumber, HEX); Serial.print(" "); Serial.println(verifyResult);

    // RSSI and SNR of the last received packet
    rssi = radio.getRSSI();
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    snr = radio.getSNR();
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");

    display.clearDisplay();
    display.setCursor(0, 15); display.print("from");
    display.setCursor(64, 15); display.print(slaveuid, HEX);  
    display.setCursor(0, 31); display.print(randomNumber, HEX);
    display.setCursor(100, 31); display.print(verifyResult);    
    display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(80, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(receivedTime / 1000); display.print(" sec");
    display.setCursor(80, 63); display.print(Vbatt); display.print(" mV");   
    display.display();
    
    // write to SD
    logFile.print(receivedTime); logFile.print(",");
    logFile.print(receivedInterval); logFile.print(",");
    logFile.print(rxdata); logFile.print(",");
    logFile.printf("%08x", slaveuid); logFile.print(",");
    logFile.print(verifyResult); logFile.print(",");    
    logFile.print(Vbatt); logFile.print(",");
    logFile.print(rssi); logFile.print(",");
    logFile.println(snr);
    logFile.flush();

    delay(1000); // for display
  }
  else {
    // some other error occurred when receiving
    Serial.print("receive failed, code ");
    Serial.println(state);
    errorBlink_1(3);                        // error [[ 3 ]]                  
  }  


  // Master transmits response data to Slave  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Serial.println("Transmitting response packet ... ");

  // send data packet
  ub.Buff_u32[0] = masteruid;         // Master device UID
  verifyNumber = randomNumber;
  ub.Buff_u32[1] = verifyNumber;      // to verify
  ub.Buff_i8[8] = rssi;               // Master RSSI  [dBm]
  ub.Buff_i8[9] = snr;                // Master SN ratio  [dB]
  ub.Buff_u8[10] = receivedInterval;  // packet reserved interval [sec]


  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  state = radio.startTransmit(ub.Buff_u8, dataNum);
  digitalWrite(LED_RED, HIGH);

  // print the packet
  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ub.Buff_u8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.printf("Master UID:\t%08x\n", masteruid);
  Serial.print("RSSI\t\t"); Serial.print(rssi); Serial.println(" dBm");
  Serial.print("SNR\t\t"); Serial.print(snr); Serial.println(" dB");
  Serial.print("verifyNumber:\t"); Serial.println(verifyNumber, HEX);

  display.clearDisplay();
  display.setCursor(0, 15); display.print("Master");
  display.setCursor(64, 15); display.print(masteruid, HEX);  
  display.setCursor(0, 31); display.print(verifyNumber, HEX);
  display.setCursor(100, 31); display.print(verifyResult);
  display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
  display.setCursor(80, 47); display.print(snr); display.print(" dB");
  display.setCursor(0, 63); display.print(receivedInterval); display.print(" sec");
  display.setCursor(80, 63); display.print(Vbatt); display.print(" mV");
  display.display();   
   
  // wait for transmittion completion
  while(!transmittedFlag) {
    delay(1);   // wait for SPI ready
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
    Serial.println();
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }
}

//*********************************************************************************
// callback function when received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}

//*********************************************************************************
// Blinks specified number of times on error
void errorBlink(uint8_t err)
{
  while(true) {
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(300);
    }
  delay(500);
  }
}

void errorBlink_1(uint8_t err)
{
  for(int i = 0; i < err; i++) {
    digitalWrite(LED_RED, LOW);
    delay(100);
    digitalWrite(LED_RED, HIGH);
    delay(300);
  }
  delay(500);
}