// ******************************************************************************************************
// BSP : STM32 MCU based boards by STMicroelectronics 2.7.1
// Borad select : STM32 based boards / LoRa boars
// Board part number : LoRa-E5 mini (Board to be selected from Tools submenu 'Board part number')
// ******************************************************************************************************
// 1. transmit data 11 bytes(SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
// 2. wait for response from Master
// 3. receive data 11 bytes(MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
// 4. verify sent randomNumber with received verifyNumber

// 2024/07/07 @msfujino

#include <Arduino.h>
#include <Adafruit_GFX.h>       // 1.11.9
#include <Adafruit_SSD1306.h>   // 2.5.10
#include "myFonts/tahoma7pt7b.h"  // Any font
#include <RadioLib.h>           // 6.6.0  RadioLib by Jan Gromes (Library manager)

#define CHECK_INTERVAL  15000   // communication check interval [mS]
#define LED_ERROR     D10       // verify error RED
#define LED_GOOD       D9       // verify good GREEN
// BAT_pad --> 1000k --> A4_pin --> 510k --> D0_pin
#define PIN_VBATT      A4       // battery voltage measurement ADC pin
#define READ_ENABLE    D0       // read PIN_VBATT enable LOW:enable, HIGH:disable

// RM0461 Reference manual Rev 8  37.1.4 IEEE 64-bit unique device ID register (UID64)
// Serial.println(*((uint32_t*)0x1FFF7580));
#define SLAVE_UID   *((uint32_t*)0x1FFF7580)    // use uid64 for Slave device UID
#define MASTER_UID  0x05358291  // example Master device UID

Adafruit_SSD1306 display(128, 64, &Wire, -1);   // OLED display

// \Arduino15\packages\STMicroelectronics\hardware\stm32\2.7.1
// \variants\STM32WLxx\WL54JCI_WL55JCI_WLE4J(8-B-C)I_WLE5J(8-B-C)I\variant_LORA_E5_MINI.h : 173
STM32WLx radio = new STM32WLx_Module();
static const uint32_t rfswitch_pins[] = {PA4, PA5, RADIOLIB_NC, RADIOLIB_NC, RADIOLIB_NC};
static const Module::RfSwitchMode_t rfswitch_table[] = {
  {STM32WLx::MODE_IDLE,  {LOW,  LOW}},
  {STM32WLx::MODE_RX,    {HIGH, LOW}},
//  {STM32WLx::MODE_TX_HP, {LOW, HIGH}},  // for LoRa-E5 mini
  {STM32WLx::MODE_TX_LP, {HIGH, HIGH}},   // for LoRa-E5-LE mini  
  END_OF_MODE_TABLE,
};

#define dataNum   11                // transmit data 11 bytes (SlaveUID 4, randomNumber 4, Vbatt 2, (previous)veryfiResult 1)
                                    // receive data 11 bytes  (MasterUID 4, verifyNumber 4, rssi 1, snr 1, interval 1)
union unionBuff {                   // buffer for data type conversion
  uint32_t  Buff_u32[dataNum/4];    // SlaveID, MasterID, randomNumber, verifyNumber
  uint16_t  Buff_u16[dataNum/2];    // Vbatt
  int8_t    Buff_i8[dataNum];       // rssi, snr
  uint8_t   Buff_u8[dataNum];       // verifyResult, interval
};
union unionBuff ub;

int8_t rssi;                        // signal RSSI [dBm]
int8_t snr;                         // signal SN ratio [dB]
uint32_t slaveuid = SLAVE_UID;      // Slave device UID
uint32_t masteruid;                 // Master device UID
uint16_t Vbatt;                        // battery voltage data [V]
char printBuff[4];                  // for sprintf()
String  txdata = "";                // transmission data packet string
String  rxdata = "";                // received data packet string
bool rxstatus = false;              // receive status flag
bool receivedFlag = false;          // flag that a packet was received
bool transmittedFlag = false;       // flag that a packet was transmitted
uint8_t receivedInterval;           // Master reveived interval [sec]
uint32_t randomNumber;              // random number to verify that data was sent correctly
uint32_t verifyNumber;              // verify number from Master
uint32_t verifyResult = 0;          // verify result 1:GOOD or 0:ERROR

int8_t txpower = 13;

// *******************************************************************************************************
void setup() {
  Serial.begin(115200);
//  while(!Serial);
  delay(2000);

  // Slave UID
  Serial.print("Slave UID "); Serial.println(*((uint32_t*)0x1FFF7580), HEX);
 
  pinMode(LED_RED, OUTPUT);         // built-in RED LED
  pinMode(LED_ERROR, OUTPUT);       // ERROR RED
  pinMode(LED_GOOD, OUTPUT);        // GOOD GREEN
  pinMode(READ_ENABLE, OUTPUT);     // read PIN_VBATT enable
  digitalWrite(LED_RED, HIGH);      // LED off
  digitalWrite(LED_ERROR, HIGH);
  digitalWrite(LED_GOOD, HIGH);  
  digitalWrite(READ_ENABLE, LOW);   // read enable

  // ADC initialization
  analogReadResolution(12);         // ADC resolution 12 bits

  // Display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Slave");  
  display.display();
  delay(2000);
  display.clearDisplay();  


  // Radio initialization
  radio.setRfSwitchTable(rfswitch_pins, rfswitch_table);

  // int16_t begin(float freq = 434.0, float bw = 125.0, uint8_t sf = 9, uint8_t cr = 7, \
  // uint8_t syncWord = RADIOLIB_SX126X_SYNC_WORD_PRIVATE, int8_t power = 10, \
  // uint16_t preambleLength = 8, float tcxoVoltage = 1.6, bool useRegulatorLDO = false);
  int state = radio.begin(922.6, 125.0, 12, 5, 0x12, 13, 8, 1.7, 0);

  if (state == RADIOLIB_ERR_NONE) {
    Serial.println("radio.begin() success!");
  }
  else {
    Serial.print("failed, code ");
    Serial.println(state);
    errorBlink(4);                // error [[ 4 ]]
  }

  // callback function when received or transmitted a packet
  radio.setDio1Action(setFlag);

  // setup end
  display.setCursor(0, 15);
  display.print("Data Verify");     
  display.setCursor(0, 47);
  display.print("LoRa-E5 Slave");
  display.setCursor(0, 63);
  display.print("Setup END");
  display.display();
  delay(2000);
  display.clearDisplay();
}

// ***********************************************************************************************************
void loop() 
{


  radio.setOutputPower(txpower);

  // ++++++++++++++++++++++++++++++++ Slave transmits a packet to Master ++++++++++++++++++++++++
  Serial.println("Transmitting packet ... ");

  // battery voltage measurement
  int n = 256;     // average times
  uint32_t Vadc = 0;
  for(int i = 0; i < n; i++) {
    Vadc = Vadc + analogRead(PIN_VBATT);
  }
  Vadc = Vadc / n;

  // example battery voltage map()data
  uint16_t adc_low = 1449;     // ADC data when battery voltage is low
  uint16_t vbatt_low = 3426;   // low Vbatt [mV]
  uint16_t adc_high = 1623;    // ADC data when battery voltage is high
  uint16_t vbatt_high = 4176;  // high Vbatt [mV]
  Vbatt = map(Vadc, adc_low, adc_high, vbatt_low, vbatt_high);   // Converts to battery voltage 3400 ~ 4200mV

  // Slave transmits data
  ub.Buff_u32[0] = slaveuid;                      // Slave device UID
  randomNumber = random(0x7FFFFFFF);              // for verification
  ub.Buff_u32[1] = randomNumber;
  ub.Buff_u16[4] = Vbatt;                         // battery voltage
  ub.Buff_u8[10] = verifyResult;                  // previous verify result

  uint16_t sec = millis() / 1000;                 // transmission time [sec]
  digitalWrite(LED_RED, LOW);
  transmittedFlag = false;
  int state = radio.startTransmit(ub.Buff_u8, dataNum);
  digitalWrite(LED_RED, HIGH);

  txdata = "";
  for(int i = 0; i < dataNum; i++) {
    sprintf(printBuff, "%02x", ub.Buff_u8[i]);
    txdata += printBuff;
  }
  txdata.toUpperCase();
  Serial.println(txdata);

  Serial.print("Slave UID:\t"); Serial.println(slaveuid, HEX);  
  Serial.print("Time:\t\t"); Serial.print(sec); Serial.println(" sec");
  Serial.print("Vbatt:\t\t"); Serial.print(Vbatt); Serial.println(" mV");
  Serial.print("randomNumber:\t"); Serial.println(randomNumber, HEX);
  Serial.print("VerifyResult:\t"); Serial.println(verifyResult);

  display.clearDisplay();
  display.setCursor(0, 15); display.print("Slave");
  display.setCursor(64, 15); display.print(slaveuid, HEX);  
  display.setCursor(0, 31); display.print(randomNumber, HEX);
  display.setCursor(100, 31); display.print(verifyResult);
  display.setCursor(0, 63); display.print(sec); display.print(" sec");
  display.setCursor(80, 63); display.print(Vbatt); display.print(" mV");   
  display.display();

  // wait for transmittion completion
  while(!transmittedFlag) {
    delay(1);   // wait for SPI ready
  }

  // error status check
  if (state == RADIOLIB_ERR_NONE) {
    // packet was successfully sent
    Serial.println("transmission finished!");
  } 
  else {
    // some other error occurred
    Serial.print("failed, code ");
    Serial.println(state);
  }

  // ++++++++++++++++++++++++++++++++Slave receives response data from Master +++++++++++++++++++++++++++
  Serial.println("Waiting for incoming Master response ... ");

  // start listening for LoRa packets
  receivedFlag = false;
  state = radio.startReceive();
  if (state != RADIOLIB_ERR_NONE) {
    Serial.print("failed code ");
    Serial.println(state);
    errorBlink_1(2);                        // error [[ 2 ]]
  }

  // check if the flag is set
  // If not received, proceed to next communication cycle to atempt synchronization recovery
  uint32_t checkstart = millis();
  digitalWrite(LED_RED, LOW);
  while(!receivedFlag & (millis() - checkstart < CHECK_INTERVAL)) {
    delay(1);     // wait for SPI ready
  }
  digitalWrite(LED_RED, HIGH);
  receivedFlag = false;

  // read a packet 
  state = radio.readData(ub.Buff_u8, dataNum);

  masteruid = ub.Buff_u32[0];   // Maste device UID
  if (state == RADIOLIB_ERR_NONE & masteruid == MASTER_UID) {   // Check Master ID 
    // packet from Master was successfully received
    rxdata = "";
    for(int i = 0; i < dataNum; i++) {
      sprintf(printBuff, "%02x", ub.Buff_u8[i]);
      rxdata += printBuff;
    }
    rxdata.toUpperCase();
    Serial.println(rxdata);

    // restoration of received data
    verifyNumber = ub.Buff_u32[1];
    rssi = ub.Buff_i8[8];  // RSSI   min. -127[dBm]
    snr = ub.Buff_i8[9];   // SNR
    receivedInterval = ub.Buff_u8[10];


    if(randomNumber != verifyNumber) {
      errorBlink_1(5);                        // error [[ 5 ]]
      verifyResult = 0;
      digitalWrite(LED_ERROR, LOW);   // RED LED on    
    } else {
      verifyResult = 1;
      digitalWrite(LED_GOOD, LOW);    // GREEN LED on
    }
    
    Serial.print("Master UID:\t"); Serial.println(masteruid, HEX);
    Serial.print("Interval:\t"); Serial.print(receivedInterval); Serial.println(" sec");
    Serial.print("RSSI:\t\t"); Serial.print(rssi); Serial.println(" dBm");
    Serial.print("SNR:\t\t"); Serial.print(snr); Serial.println(" dB");
    Serial.print("Vbatt:\t\t"); Serial.print(Vbatt); Serial.println(" mV");
    Serial.print("verifyNumber:\t"); Serial.print(verifyNumber, HEX); Serial.print(" "); Serial.println(verifyResult);

    display.clearDisplay();
    display.setCursor(0, 15); display.print("from");
    display.setCursor(56, 15); display.print(masteruid, HEX);  
    display.setCursor(0, 31); display.print(verifyNumber, HEX);
    display.setCursor(100, 31); display.print(verifyResult);
    display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(80, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(receivedInterval); display.print(" sec");
    display.setCursor(80, 63); display.print(Vbatt); display.print(" mV");   
    display.display();
  }
  // error status check
  else {
    // some other error occurred when receiving
    Serial.print("receive failed, code ");
    Serial.println(state);
    errorBlink_1(3);                        // error [[ 3 ]]
  }  
  delay(1500);

  // LEDs off
  digitalWrite(LED_ERROR, HIGH); 
  digitalWrite(LED_GOOD, HIGH);
}

//****************************************************************************************
// this function is called when a complete packet is received or transmitted
void setFlag(void) 
{
  uint16_t irqstatus = radio.getIrqStatus();
  if(irqstatus == RADIOLIB_SX126X_IRQ_RX_DONE) {
    receivedFlag = true;
  }
  else if(irqstatus == RADIOLIB_SX126X_IRQ_TX_DONE) {
    transmittedFlag = true;
  }
  else {
    receivedFlag = false;
    transmittedFlag = false;
  }
}

//****************************************************************************************
// Blink specified number of times on error
void errorBlink(uint8_t err)
{
  while(true) {
    for(int i = 0; i < err; i++) {
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(300);
    }
    delay(500);
  }
}

void errorBlink_1(uint8_t err)
{
  for(int i = 0; i < err; i++) {
    digitalWrite(LED_RED, LOW);
    delay(100);
    digitalWrite(LED_RED, HIGH);
    delay(300);
  }
  delay(500);
}